/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.vorgang;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.*;

import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.RepresentationModel;
import org.springframework.hateoas.mediatype.hal.HalModelBuilder;
import org.springframework.hateoas.server.RepresentationModelAssembler;
import org.springframework.stereotype.Component;
import org.springframework.web.util.UriComponentsBuilder;

import de.ozgcloud.alfa.common.ModelBuilder;
import de.ozgcloud.alfa.common.command.CommandController.CommandByRelationController;
import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserId;
import de.ozgcloud.alfa.common.user.UserRole;
import de.ozgcloud.alfa.statistic.Statistic;
import de.ozgcloud.alfa.wiedervorlage.WiedervorlageController;
import lombok.RequiredArgsConstructor;

@Component
@RequiredArgsConstructor
class VorgangModelAssembler implements RepresentationModelAssembler<Vorgang, EntityModel<Vorgang>> {

	static final String REL_NEXT = "next";
	static final String REL_PREVIOUS = "previous";
	static final String REL_VORGANG_MIT_EINGANG = "vorgang-mit-eingang";
	static final String REL_WIEDERVORLAGEN = "wiedervorlagen";
	static final String REL_VORGANG_ASSIGN = "assign";

	private final CurrentUserService userService;
	private final VorgangService vorgangService;

	public RepresentationModel<EntityModel<EnhancedVorgang>> toCollectionModel(Stream<VorgangHeader> entities, VorgaengeHeaderResponse response,
			FindVorgaengeHeaderRequestCriteria requestCriteria, Statistic statistic) {
		var model = initModel(entities, statistic);

		getPreviousPage(requestCriteria.getPage()).map(page -> buildPageLink(page, REL_PREVIOUS, requestCriteria)).ifPresent(model::add);
		getNextPage(response, requestCriteria).map(page -> buildPageLink(page, REL_NEXT, requestCriteria)).ifPresent(model::add);

		return model;
	}

	RepresentationModel<EntityModel<EnhancedVorgang>> initModel(Stream<VorgangHeader> entities, Statistic statistic) {
		return HalModelBuilder.halModelOf(EnhancedVorgang.builder().statistic(statistic).build())
				.link(linkTo(VorgangController.class).withSelfRel())
				.embed(entities.map(this::toModel).toList())
				.build();
	}

	Optional<Integer> getPreviousPage(int currentPage) {
		return isFirstPage(currentPage) ? Optional.empty() : Optional.of(currentPage - 1);
	}

	Optional<Integer> getNextPage(VorgaengeHeaderResponse response, FindVorgaengeHeaderRequestCriteria requestCriteria) {
		return Optional.of(getLastEntryIndex(requestCriteria, response.getVorgaengeHeader().size()))
				.filter(index -> response.getTotal() > index)
				.map(index -> requestCriteria.getPage() + 1);
	}

	int getLastEntryIndex(FindVorgaengeHeaderRequestCriteria requestCriteria, int currentPageSize) {
		if (isFirstPage(requestCriteria.getPage())) {
			return currentPageSize;
		}
		return currentPageSize + (requestCriteria.getPage() * requestCriteria.getLimit());
	}

	private boolean isFirstPage(int currentPage) {
		return currentPage == 0;
	}

	Link buildPageLink(int page, String linkRel, FindVorgaengeHeaderRequestCriteria requestCriteria) {
		var uriBuilder = createVorgangUriBuilder(page, requestCriteria);

		requestCriteria.getSearchBy().ifPresent(searchBy -> uriBuilder.queryParam(VorgangController.PARAM_SEARCH, searchBy));
		handleAssignedTo(uriBuilder, requestCriteria);
		requestCriteria.getFilterCriteria().getStatus().ifPresent(status -> uriBuilder.queryParam(VorgangController.PARAM_STATUS, status));
		handleNextWiedevorlageFrist(uriBuilder, requestCriteria);

		return Link.of(uriBuilder.build().toUri().toASCIIString()).withRel(linkRel);
	}

	private void handleAssignedTo(UriComponentsBuilder uriBuilder, FindVorgaengeHeaderRequestCriteria requestCriteria) {
		if (requestCriteria.getFilterCriteria().isFilterByAssignedTo()) {
			uriBuilder.queryParam(VorgangController.PARAM_ASSIGNED_TO, getAssignedToValue(requestCriteria.getFilterCriteria()));
		}
	}

	private void handleNextWiedevorlageFrist(UriComponentsBuilder uriBuilder, FindVorgaengeHeaderRequestCriteria requestCriteria) {
		if (requestCriteria.getFilterCriteria().isHasNextWiedervorlageFrist()) {
			uriBuilder.queryParam(VorgangController.PARAM_NEXT_WIEDERVORLAGE_FRIST, VorgangController.PARAM_NEXT_WIEDERVORLAGE_FRIST_EXISTS);
		}
	}

	UriComponentsBuilder createVorgangUriBuilder(int page, FindVorgaengeHeaderRequestCriteria requestCriteria) {
		return linkTo(VorgangController.class).toUriComponentsBuilder()
				.queryParam(VorgangController.PARAM_PAGE, page)
				.queryParam(VorgangController.PARAM_LIMIT, requestCriteria.getLimit());
	}

	String getAssignedToValue(FilterCriteria filterCriteria) {
		return filterCriteria.getAssignedTo().map(UserId::toString).orElse(StringUtils.EMPTY);
	}

	@Override
	public EntityModel<Vorgang> toModel(Vorgang vorgang) {
		var selfLinkBuilder = linkTo(VorgangController.class).slash(vorgang.getId());
		var assignLink = linkTo(methodOn(CommandByRelationController.class).createCommand(vorgang.getId(), vorgang.getId(),
				vorgang.getVersion(), null)).withRel(REL_VORGANG_ASSIGN);
		var wiedervorlagenLink = linkTo(methodOn(WiedervorlageController.class).findByVorgang(vorgang.getId())).withRel(REL_WIEDERVORLAGEN);

		return ModelBuilder.fromEntity(vorgang)
				.addLink(selfLinkBuilder.withSelfRel())
				.addLink(selfLinkBuilder.withRel(REL_VORGANG_MIT_EINGANG))
				.ifMatch(() -> isAssigneable(vorgang))
				.addLink(assignLink)
				.ifMatch(() -> userService.hasRole(UserRole.VERWALTUNG_USER)).addLink(wiedervorlagenLink)
				.buildModel();
	}

	boolean isAssigneable(Vorgang vorgang) {
		return (userService.hasRole(UserRole.VERWALTUNG_USER) || userService.hasRole(UserRole.VERWALTUNG_POSTSTELLE))
				&& vorgangService.isVorgangEditable(vorgang);
	}
}