/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.historie;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandService;
import de.ozgcloud.alfa.common.command.CommandTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;

class HistorieServiceTest {

	@Spy
	@InjectMocks
	private HistorieService service;
	@Mock
	private CommandService commandService;
	@Spy
	private List<HistorieProcessor> processors = new ArrayList<>();
	@Mock
	private HistorieProcessor firstProcessor;
	@Mock
	private HistorieProcessor secondProcessor;

	@Nested
	class TestFindFinishedCommands {

		private final Stream<Command> commands = Stream.of(CommandTestFactory.create());

		@BeforeEach
		void setUpMock() {
			when(commandService.findFinishedCommands(any())).thenReturn(commands);
		}

		@Test
		void shouldCallCommandServiceForFinishedCommands() {
			findFinishedCommands();

			verify(commandService).findFinishedCommands(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldCallProcessCommands() {
			findFinishedCommands();

			verify(service).processCommands(commands);
		}

		@Test
		void shouldReturnProcessedCommands() {
			var processedCommands = Stream.of(CommandTestFactory.create());
			doReturn(processedCommands).when(service).processCommands(commands);

			var finishedCommands = findFinishedCommands();

			assertThat(finishedCommands).isSameAs(processedCommands);
		}

		private Stream<Command> findFinishedCommands() {
			return service.findFinishedCommands(VorgangHeaderTestFactory.ID);
		}
	}

	@Nested
	class TestProcessCommands {

		private final Stream<Command> inputCommands = Stream.of(CommandTestFactory.create());
		private final Stream<Command> processedOnceCommands = Stream.of(CommandTestFactory.create());
		private final Stream<Command> processedTwiceCommands = Stream.of(CommandTestFactory.create());

		@BeforeEach
		void setProcessorMocks() {
			processors.addAll(List.of(firstProcessor, secondProcessor));
			when(firstProcessor.process(any())).thenReturn(processedOnceCommands);
			when(secondProcessor.process(any())).thenReturn(processedTwiceCommands);
		}

		@Test
		void shouldProcessCommands() {
			service.processCommands(inputCommands);

			verify(firstProcessor).process(inputCommands);
		}

		@Test
		void shouldDoSubsequentProcessing() {
			service.processCommands(inputCommands);

			verify(secondProcessor).process(processedOnceCommands);
		}

		@Test
		void shouldReturnCommandsAfterProcessing() {
			var resultCommands = service.processCommands(inputCommands);

			assertThat(resultCommands).isEqualTo(processedTwiceCommands);
		}
	}
}