/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.vorgang;

import static de.ozgcloud.alfa.vorgang.VorgangRemoteService.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.alfa.common.user.UserProfileTestFactory;
import de.ozgcloud.vorgang.common.GrpcQueryOperator;
import de.ozgcloud.vorgang.vorgang.GrpcFilterBy;
import de.ozgcloud.vorgang.vorgang.GrpcFindVorgangRequest;
import de.ozgcloud.vorgang.vorgang.GrpcFindVorgangResponse;
import de.ozgcloud.vorgang.vorgang.GrpcQuery;
import de.ozgcloud.vorgang.vorgang.GrpcVorgangQueryExpression;
import de.ozgcloud.vorgang.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;

class VorgangRemoteServiceTest {

	@Spy
	@InjectMocks // NOSONAR
	private VorgangRemoteService service;

	@Mock
	private VorgangServiceBlockingStub serviceStub;
	@Mock
	private VorgangHeaderMapper vorgangHeaderMapper;
	@Mock
	private VorgangWithEingangMapper vorgangWithEingangMapper;

	private final GrpcFilterBy filterBy = GrpcFilterByTestFactory.create();

	private final GrpcQuery query = GrpcQueryTestFactory.create();

	@DisplayName("Find vorgaenge header")
	@Nested
	class TestFindVorgaengeHeader {

		private final FilterCriteria filterCriteria = FilterCriteriaTestFactory.create();

		@DisplayName("should")
		@Nested
		class TestProzess {

			private final FindVorgaengeHeaderRequestCriteria requestCriteria = FindVorgaengeRequestCriteriaTestFactory.create();

			@BeforeEach
			void initMockReturnValues() {
				doReturn(query).when(service).createQuery(any());
				doReturn(filterBy).when(service).createFilterBy(any());
				doReturn(VorgangListResponseTestFactory.create()).when(service).buildVorgaengeHeaderResponse(any());
			}

			@Test
			void callBuildRequest() {
				callService();

				verify(service).buildFindVorgangRequest(requestCriteria, filterCriteria);
			}

			@Test
			void callCreateQuery() {
				callService();

				verify(service).createQuery(filterCriteria);
			}

			@Test
			void callCreateFilterBy() {
				callService();

				verify(service).createFilterBy(filterCriteria);
			}

			@Test
			void callStub() {
				callService();

				verify(serviceStub).findVorgang(any());
			}

			private VorgaengeHeaderResponse callService() {
				return service.findVorgaengeHeader(requestCriteria, filterCriteria);
			}
		}

		@DisplayName("build find vorgang request")
		@Nested
		class TestBuildFindVorgangRequest {

			@BeforeEach
			void initMockReturnValues() {
				doReturn(query).when(service).createQuery(any());
				doReturn(filterBy).when(service).createFilterBy(any());
			}

			@Test
			void shouldSetLimit() {
				var request = callService();

				assertThat(request.getLimit()).isEqualTo(FindVorgaengeRequestCriteriaTestFactory.LIMIT);
			}

			@Test
			void shouldSetOffset() {
				var request = callService();

				assertThat(request.getOffset())
						.isEqualTo(FindVorgaengeRequestCriteriaTestFactory.PAGE * FindVorgaengeRequestCriteriaTestFactory.LIMIT);
			}

			@Test
			void shouldSetQuery() {
				var request = callService();

				assertThat(request.getQuery()).isEqualTo(query);
			}

			@Test
			void shouldSetFilterBy() {
				var request = callService();

				assertThat(request.getFilterBy()).isEqualTo(filterBy);
			}

			@Test
			void shouldSetOrderBy() {
				var request = callService();

				assertThat(request.getOrderBy().name()).isEqualTo(FindVorgaengeRequestCriteriaTestFactory.ORDER_BY.name());
			}

			private GrpcFindVorgangRequest callService() {
				return service.buildFindVorgangRequest(FindVorgaengeRequestCriteriaTestFactory.create(), filterCriteria);
			}
		}

		@DisplayName("response")
		@Nested
		class TestResponse {

			private final GrpcFindVorgangResponse findVorgangRespone = GrpcVorgangTestFactory.createVorgangResponse();
			private final VorgangHeader vorgangHeader = VorgangHeaderTestFactory.create();

			@BeforeEach
			void initMockReturnValues() {
				when(vorgangHeaderMapper.toVorgangHeader(any())).thenReturn(vorgangHeader);
			}

			@Test
			void shouldReturnMappedVorgangHeader() {
				var vorgangResponse = callService();

				assertThat(vorgangResponse.getVorgaengeHeader()).hasSize(1);
				assertThat(vorgangResponse.getVorgaengeHeader().get(0)).isEqualTo(vorgangHeader);
			}

			@Test
			void shouldReturnTotal() {
				var vorgangResponse = callService();

				assertThat(vorgangResponse.getTotal()).isEqualTo(findVorgangRespone.getTotal());
			}

			@Test
			void shouldMapVorgangHeader() {
				callService();

				verify(vorgangHeaderMapper).toVorgangHeader(any());
			}

			private VorgaengeHeaderResponse callService() {
				return service.buildVorgaengeHeaderResponse(findVorgangRespone);
			}
		}
	}

	@Nested
	class TestCreateQuery {

		@Nested
		class WhenHasUngeleseneNachrichtenIsTrue {

			private final FilterCriteria filterCriteria = FilterCriteriaTestFactory.createBuilder().hasUngeleseneNachrichten(true).build();

			@Test
			void shouldAddExpression() {
				var query = service.createQuery(filterCriteria);

				assertThat(query.getExpressionsList()).extracting(GrpcVorgangQueryExpression::getPath)
						.containsOnlyOnce(VORGAENGE_WITH_UNGELESENE_NACHRICHTEN_PATH);
			}

			@Test
			void shouldHaveOperator() {
				var expression = getExpression(service.createQuery(filterCriteria));

				assertThat(expression.getOperator()).isEqualTo(GrpcQueryOperator.EQUAL);
			}

			@Test
			void shouldHaveOperandBoolValue() {
				var expression = getExpression(service.createQuery(filterCriteria));

				assertThat(expression.getOperandBoolValue()).isTrue();
			}

			private GrpcVorgangQueryExpression getExpression(GrpcQuery query) {
				return query.getExpressionsList().stream()
						.filter(expression -> expression.getPath().equals(VORGAENGE_WITH_UNGELESENE_NACHRICHTEN_PATH)).findFirst().orElseThrow();
			}
		}

		@Nested
		class WhenHasUngeleseneNachrichtenIsFalse {

			private final FilterCriteria filterCriteria = FilterCriteriaTestFactory.createBuilder().hasUngeleseneNachrichten(false).build();

			@Test
			void shouldNotAddExpression() {
				var query = service.createQuery(filterCriteria);

				assertThat(query.getExpressionsList()).extracting(GrpcVorgangQueryExpression::getPath)
						.isNotEqualTo(VORGAENGE_WITH_UNGELESENE_NACHRICHTEN_PATH);
			}
		}
	}

	@DisplayName("Create filterBy")
	@Nested
	class TestCreateFilterBy {

		@Test
		void shouldSetFilterByOrganisationsEinheitId() {
			var filterCriteria = createFilterBy();

			assertThat(filterCriteria.getFilterByOrganisationseinheitenId())
					.isEqualTo(FilterCriteriaTestFactory.FILTER_BY_ORGANISATIONS_EINHEITEN_ID);
		}

		@Test
		void shouldSetOrganisationsEinheitIds() {
			var filterCriteria = createFilterBy();

			assertThat(filterCriteria.getOrganisationseinheitIdList()).containsExactly(FilterCriteriaTestFactory.ORGANISATIONS_EINHEIT_ID);
		}

		@DisplayName("filterAssignedTo")
		@Nested
		class TestFilterAssignedTo {

			@Test
			void shouldBeSet() {
				var filterCriteria = createFilterBy();

				assertThat(filterCriteria.getFilterByAssignedTo()).isTrue();
			}
		}

		@DisplayName("assignedTo")
		@Nested
		class TestAssignedTo {

			@Test
			void shouldBeSetIfExists() {
				var filterCriteria = createFilterBy();

				assertThat(filterCriteria.getAssignedTo()).isEqualTo(UserProfileTestFactory.ID.toString());
			}

			@Test
			void shouldNotBeSetIfNotExist() {
				var filterCriteria = createFilterBy(FilterCriteriaTestFactory.createBuilder().assignedTo(Optional.empty()).build());

				assertThat(filterCriteria.getAssignedTo()).isEqualTo(StringUtils.EMPTY);
			}
		}

		@DisplayName("status")
		@Nested
		class TestStatus {

			@Test
			void shouldBeSetIfExists() {
				var filterCriteria = createFilterBy();

				assertThat(filterCriteria.getStatusList()).containsExactly(FilterCriteriaTestFactory.VORGANG_STATUS.toString());
			}

			@Test
			void shouldNotBeSetIfNotExist() {
				var filterCriteria = createFilterBy(FilterCriteriaTestFactory.createBuilder().status(Optional.empty()).build());

				assertThat(filterCriteria.getStatusCount()).isZero();

			}
		}

		@DisplayName("hasNextWiedervorlageFrist")
		@Nested
		class TestHasNextWiedervorlageFrist {

			@Test
			void shouldBeSetToFalse() {
				var filterCriteria = createFilterBy(FilterCriteriaTestFactory.createBuilder().hasNextWiedervorlageFrist(false).build());

				assertThat(filterCriteria.getHasNextWiedervorlageFrist()).isFalse();
			}

			@Test
			void shouldNotBeSetToTrue() {
				var filterCriteria = createFilterBy(FilterCriteriaTestFactory.createBuilder().hasNextWiedervorlageFrist(true).build());

				assertThat(filterCriteria.getHasNextWiedervorlageFrist()).isTrue();
			}
		}

		private GrpcFilterBy createFilterBy() {
			return createFilterBy(FilterCriteriaTestFactory.create());
		}

		private GrpcFilterBy createFilterBy(FilterCriteria filterCriteria) {
			return service.createFilterBy(filterCriteria);
		}
	}

	@DisplayName("Find vorgangWithEingang")
	@Nested
	class TestFindVorgangWithEingang {

		private final FilterCriteria filterCriteria = FilterCriteriaTestFactory.create();

		@Nested
		class TestCalls {

			private final VorgangWithEingang vorgangWithEingang = VorgangWithEingangTestFactory.create();

			@BeforeEach
			void mockReturnValue() {
				when(serviceStub.findVorgangWithEingang(any()))
						.thenReturn(GrpcVorgangWithEingangResponseTestFactory.createVorgangWithEingangResponse());
				when(vorgangWithEingangMapper.toVorgangWithEingang(any())).thenReturn(vorgangWithEingang);
				doReturn(filterBy).when(service).createFilterBy(any());
			}

			@Test
			void shouldCreateFilterBy() {
				callService();

				verify(service).createFilterBy(any());
			}

			@Test
			void shouldCallBuildRequest() {
				callService();

				verify(service).buildFindVorgangWithEingangRequest(VorgangHeaderTestFactory.ID, filterCriteria);
			}

			@Test
			void shouldCallStub() {
				callService();

				verify(serviceStub).findVorgangWithEingang(any());
			}

			@Test
			void shouldCallMapper() {
				callService();

				verify(vorgangWithEingangMapper).toVorgangWithEingang(any());
			}

			@Test
			void shouldReturn() {
				var result = callService();

				assertThat(result).isEqualTo(vorgangWithEingang);
			}
		}

		@DisplayName("build request")
		@Nested
		class TestBuildRequest {

			@BeforeEach
			void mockReturnValue() {
				doReturn(filterBy).when(service).createFilterBy(any());
			}

			@Test
			void shouldHaveFilter() {
				var filter = service.buildFindVorgangWithEingangRequest(VorgangHeaderTestFactory.ID, filterCriteria);

				assertThat(filter.getFilterBy()).isNotNull();
			}

			@Test
			void shouldContainsVorgangId() {
				var filter = service.buildFindVorgangWithEingangRequest(VorgangHeaderTestFactory.ID, filterCriteria);

				assertThat(filter.getId()).isEqualTo(VorgangHeaderTestFactory.ID);
			}
		}

		private VorgangWithEingang callService() {
			return service.findVorgangWithEingang(VorgangHeaderTestFactory.ID, filterCriteria);
		}
	}
}