/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.vorgang;

import static de.ozgcloud.alfa.common.UserProfileUrlProviderTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.LinkRelation;
import org.springframework.web.util.UriComponentsBuilder;
import org.springframework.web.util.UriTemplate;

import de.ozgcloud.alfa.common.UserProfileUrlProvider;
import de.ozgcloud.alfa.common.command.CommandController.CommandByRelationController;
import de.ozgcloud.alfa.common.user.UserManagerUrlProvider;
import de.ozgcloud.alfa.postfach.PostfachMailController;

class VorgangWithEingangProcessorTest {

	@Spy
	@InjectMocks
	private VorgangWithEingangProcessor processor;
	@Mock
	private PostfachMailController postfachMailController;
	@Mock
	private UserManagerUrlProvider userManagerUrlProvider;

	@Mock
	private VorgangProcessorProperties vorgangProcessorProperties;

	private final UserProfileUrlProvider urlProvider = new UserProfileUrlProvider();

	@DisplayName("Attachments")
	@Nested
	class TestAttachments {

		@BeforeEach
		void init() {
			initUserProfileUrlProvider(urlProvider);
		}

		@DisplayName("link")
		@Nested
		class TestLink {

			private static final String PATH = "/api/vorgang/" + VorgangHeaderTestFactory.ID + "/attachments";
			private final LinkRelation linkRel = VorgangWithEingangProcessor.REL_ATTACHMENTS;

			@DisplayName("should be present if attachments exists")
			@Test
			void shouldBePresent() {
				var link = processor.process(buildModelWithAttachments(1)).getLink(linkRel);

				assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(PATH);
			}

			@Test
			void shouldNotBePresent() {
				var link = processor.process(buildModelWithAttachments(0)).getLink(linkRel);

				assertThat(link).isNotPresent();
			}
		}

		@DisplayName("download link")
		@Nested
		class TestDownloadLink {

			private static final String PATH = "/api/vorgang/" + VorgangHeaderTestFactory.ID + "/attachments";
			private final LinkRelation linkRel = VorgangWithEingangProcessor.REL_DOWNLOAD_ATTACHMENTS;

			@DisplayName("should be present if attachments exists")
			@Test
			void shouldBePresent() {
				var link = processor.process(buildModelWithAttachments(1)).getLink(linkRel);

				assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(PATH);
			}

			@Test
			void shouldNotBePresent() {
				var link = processor.process(buildModelWithAttachments(0)).getLink(linkRel);

				assertThat(link).isNotPresent();
			}
		}

		private EntityModel<VorgangWithEingang> buildModelWithAttachments(int numberOfAttachments) {
			return buildModelWithEingang(EingangTestFactory.createBuilder()
					.numberOfAttachments(numberOfAttachments)
					.build());
		}

		private EntityModel<VorgangWithEingang> buildModelWithEingang(Eingang eingang) {
			return EntityModel.of(VorgangWithEingangTestFactory.createBuilder().eingang(eingang).build());
		}
	}

	@DisplayName("Representations")
	@Nested
	class TestRepresentationsLink {

		@BeforeEach
		void init() {
			initUserProfileUrlProvider(urlProvider);
		}

		private static final String PATH = "/api/vorgang/" + VorgangHeaderTestFactory.ID + "/representations";

		private final LinkRelation linkRel = VorgangWithEingangProcessor.REL_REPRESENTATIONS;

		@DisplayName("should be present if representations exist")
		@Test
		void shouldBePresent() {
			var link = processor.process(buildModelWithRepresentation(1)).getLink(linkRel);

			assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo(PATH);
		}

		@Test
		void shouldNotBePresent() {
			var link = processor.process(buildModelWithRepresentation(0)).getLink(linkRel);

			assertThat(link).isNotPresent();
		}

		private EntityModel<VorgangWithEingang> buildModelWithRepresentation(int numberOfRepresentations) {
			return buildModelWithEingang(EingangTestFactory.createBuilder()
					.numberOfRepresentations(numberOfRepresentations)
					.build());
		}

		private EntityModel<VorgangWithEingang> buildModelWithEingang(Eingang eingang) {
			return EntityModel.of(VorgangWithEingangTestFactory.createBuilder().eingang(eingang).build());
		}
	}

	@Nested
	class TestPostfachMailsLink {

		private final EntityModel<VorgangWithEingang> model = EntityModel.of(VorgangWithEingangTestFactory.create());

		@Test
		void shouldCallPostfachMailController() {
			processor.process(model).getLink(VorgangWithEingangProcessor.REL_POSTFACH_MAILS);

			verify(postfachMailController).isPostfachConfigured();
		}

		@Test
		void shouldBePresentIfConfigured() {
			when(postfachMailController.isPostfachConfigured()).thenReturn(true);

			var link = processor.process(model).getLink(VorgangWithEingangProcessor.REL_POSTFACH_MAILS);

			assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo("/api/postfachMails?vorgangId=" + VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldNotBePresentIfNotConfigured() {
			when(postfachMailController.isPostfachConfigured()).thenReturn(false);

			var link = processor.process(model).getLink(VorgangWithEingangProcessor.REL_POSTFACH_MAILS);

			assertThat(link).isEmpty();
		}
	}

	@DisplayName("Historie Link")
	@Nested
	class TestHistorieLink {

		@Test
		void shouldBePresent() {
			var link = processor.process(EntityModel.of(VorgangWithEingangTestFactory.create())).getLink(VorgangWithEingangProcessor.REL_HISTORIE);

			assertThat(link).isPresent().get().extracting(Link::getHref).isEqualTo("/api/histories?vorgangId=" + VorgangHeaderTestFactory.ID);
		}
	}

	@DisplayName("Search user profile link")
	@Nested
	class TestSearchUserProfileLink {

		@BeforeEach
		void init() {
			initUserProfileUrlProvider(urlProvider);
		}

		@DisplayName("on configured for search user profile")
		@Nested
		class TestIsConfigured {

			@BeforeEach
			void init() {
				when(userManagerUrlProvider.isConfiguredForSearchUserProfile()).thenReturn(true);
			}

			@Test
			void shouldAddIfConfigured() {
				when(userManagerUrlProvider.getSearchUserProfilesTemplate()).thenReturn(
						UriComponentsBuilder.fromUriString("/api/userProfiles/?searchBy={searchBy}"));

				var model = processor.process(EntityModel.of(VorgangWithEingangTestFactory.create()));

				assertThat(model.getLink(VorgangWithEingangProcessor.REL_SEARCH_USER)).isPresent().get().extracting(Link::getHref).isEqualTo(
						"/api/userProfiles/?searchBy={searchBy}&deleted=false&organisationsEinheitId="
								+ ZustaendigeStelleTestFactory.ORGANISATIONSEINHEITEN_ID);
			}

			@Test
			void shouldNotAddIfOrganisationsEinheitIdIsMissing() {
				var model = processor.process(EntityModel.of(createVorgangWithoutOrganisationsEinheitId()));

				assertThat(model.getLink(VorgangWithEingangProcessor.REL_SEARCH_USER)).isNotPresent();
			}
		}

		@DisplayName("on not configured for search user profile")
		@Nested
		class TestIsNotConfigured {

			@BeforeEach
			void init() {
				when(userManagerUrlProvider.isConfiguredForSearchUserProfile()).thenReturn(false);
			}

			@Test
			void shouldNotAddifNotConfigured() {
				var model = processor.process(EntityModel.of(VorgangWithEingangTestFactory.create()));

				assertThat(model.getLink(VorgangWithEingangProcessor.REL_SEARCH_USER)).isNotPresent();
			}

			@Test
			void shouldNotAddIfOrganisationsEinheitIdIsPresent() {
				var model = processor.process(EntityModel.of(createVorgangWithoutOrganisationsEinheitId()));

				assertThat(model.getLink(VorgangWithEingangProcessor.REL_SEARCH_USER)).isNotPresent();
			}
		}

		private VorgangWithEingang createVorgangWithoutOrganisationsEinheitId() {
			var zustaendigeStelle = ZustaendigeStelleTestFactory.createBuilder().organisationseinheitenId(null).build();
			var eingang = EingangTestFactory.createBuilder().zustaendigeStelle(zustaendigeStelle).build();

			return VorgangWithEingangTestFactory.createBuilder().eingang(eingang).build();
		}
	}

	@DisplayName("Process vorgang")
	@Nested
	class TestProcessVorgang {

		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();
		private final EntityModel<VorgangWithEingang> vorgangEntityModel = EntityModel.of(vorgang);

		@BeforeEach
		void init() {
			initUserProfileUrlProvider(urlProvider);
		}

		@DisplayName("on matching configuration")
		@Nested
		class TestOnMatchingConfiguration {

			@BeforeEach
			void init() {
				when(vorgangProcessorProperties.getProcessor()).thenReturn(List.of(VorgangPropertyTestFactory.create()));
			}

			@Test
			void shouldBeVisible() {
				var entityModel = processor.process(vorgangEntityModel);

				assertThat(entityModel.getLink(VorgangWithEingangProcessor.REL_PROCESS_VORGANG))
						.isPresent().get().extracting(Link::getHref).isEqualTo(buildExpectedLink());
			}

			private String buildExpectedLink() {
				return new UriTemplate(CommandByRelationController.COMMAND_BY_RELATION_PATH)
						.expand(VorgangHeaderTestFactory.ID, VorgangHeaderTestFactory.ID, VorgangHeaderTestFactory.VERSION).toString();
			}
		}

		@DisplayName("on non matching configuration")
		@Nested
		class TestOnNonMatchingConfiguration {

			@Test
			void shouldNotBeVisibleOnWrongFormId() {
				var properties = List.of(VorgangPropertyTestFactory.createBuilder().formId("quatsch").build());
				when(vorgangProcessorProperties.getProcessor()).thenReturn(properties);

				var model = processor.process(vorgangEntityModel);

				assertThat(model.getLink(VorgangWithEingangProcessor.REL_PROCESS_VORGANG)).isNotPresent();
			}

			@Test
			void shouldNotBeVisibleOnWrongFormEngineName() {
				var properties = List.of(VorgangPropertyTestFactory.createBuilder().formEngineName("quatsch").build());
				when(vorgangProcessorProperties.getProcessor()).thenReturn(properties);

				var model = processor.process(vorgangEntityModel);

				assertThat(model.getLink(VorgangWithEingangProcessor.REL_PROCESS_VORGANG)).isNotPresent();
			}
		}
	}
}