package de.ozgcloud.alfa.wiedervorlage;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Spy;
import org.mockito.stubbing.Answer;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandBodyMapper;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.CommandTestFactory;
import de.ozgcloud.alfa.common.command.LegacyOrder;
import io.jsonwebtoken.lang.Collections;

class WiedervorlageHistorieProcessorTest {

	@Spy
	private WiedervorlageHistorieProcessor processor;

	@Nested
	class TestProcess {

		private final Command command = CommandTestFactory.create();
		private final Command translatedCommand = CommandTestFactory.create();
		private final Answer<Command> doNotTranslate = invocation -> invocation.getArgument(0);

		@BeforeEach
		void mockDefault() {
			doAnswer(doNotTranslate).when(processor).translateCreateAttachedItem(any());
			doAnswer(doNotTranslate).when(processor).translateUpdateAttachedItem(any());
			doAnswer(doNotTranslate).when(processor).translatePatchAttachedItem(any());
		}

		@Test
		void shouldPropagateAllOrders() {

			var commands = processor.process(Stream.of(command));

			assertThat(commands).containsExactly(command);
		}

		@Test
		void shouldTranslateCreateAttachedItem() {
			processor.process(Stream.of(command)).toList();

			verify(processor).translateCreateAttachedItem(command);
		}

		@Test
		void shouldReturnTranslatedCreateAttachedItem() {
			doReturn(translatedCommand).when(processor).translateCreateAttachedItem(any());

			var commands = processor.process(Stream.of(command)).toList();

			assertThat(commands).containsExactly(translatedCommand);
			reset(processor);
		}

		@Test
		void shouldTranslateUpdateAttachedItem() {
			processor.process(Stream.of(command)).toList();

			verify(processor).translateUpdateAttachedItem(command);
		}

		@Test
		void shouldReturnTranslatedUpdateAttachedItem() {
			doReturn(translatedCommand).when(processor).translateUpdateAttachedItem(any());

			var commands = processor.process(Stream.of(command)).toList();

			assertThat(commands).containsExactly(translatedCommand);
			reset(processor);
		}

		@Test
		void shouldTranslatePatchAttachedItem() {
			processor.process(Stream.of(command)).toList();

			verify(processor).translatePatchAttachedItem(command);
		}

		@Test
		void shouldReturnTranslatedPatchAttachedItem() {
			doReturn(translatedCommand).when(processor).translatePatchAttachedItem(any());

			var commands = processor.process(Stream.of(command)).toList();

			assertThat(commands).containsExactly(translatedCommand);
			reset(processor);
		}
	}

	@Nested
	class TestTranslateCreateAttachedItem {

		private final Map<String, ?> commandBody = Map.of(CommandBodyMapper.ITEM_NAME_PROPERTY, Wiedervorlage.class.getSimpleName());
		private final Command createWiedervorlageCommand = CommandTestFactory.createBuilder()
				.order(CommandOrder.CREATE_ATTACHED_ITEM.name())
				.body(commandBody)
				.build();

		@Test
		void shouldTranslateOrderToCreateWiedervorlage() {
			var expectedCommand = createWiedervorlageCommand.toBuilder().order(LegacyOrder.CREATE_WIEDERVORLAGE).build();

			var translatedCommand = processor.translateCreateAttachedItem(createWiedervorlageCommand);

			assertThat(translatedCommand).usingRecursiveComparison().isEqualTo(expectedCommand);
		}

		@Test
		void shouldNotTranslateOtherOrderCommands() {
			var command = createWiedervorlageCommand.toBuilder().order(LoremIpsum.getInstance().getWords(1)).build();

			var translatedCommand = processor.translateCreateAttachedItem(command);

			assertThat(translatedCommand).isSameAs(command);
		}

		@Test
		void shouldNotTranslateOtherAttachedItems() {
			var otherCommandBody = Map.of(CommandBodyMapper.ITEM_NAME_PROPERTY, LoremIpsum.getInstance().getWords(1));
			var command = createWiedervorlageCommand.toBuilder().body(otherCommandBody).build();

			var translatedCommand = processor.translateCreateAttachedItem(command);

			assertThat(translatedCommand).isSameAs(command);
		}
	}

	@Nested
	class TestTranslateUpdateAttachedItem {

		private final Map<String, ?> commandBody = Map.of(CommandBodyMapper.ITEM_NAME_PROPERTY, Wiedervorlage.class.getSimpleName());
		private final Command editWiedervorlageCommand = CommandTestFactory.createBuilder()
				.order(CommandOrder.UPDATE_ATTACHED_ITEM.name())
				.body(commandBody)
				.build();

		@Test
		void shouldTranslateOrderToEditWiedervorlage() {
			var expectedCommand = editWiedervorlageCommand.toBuilder().order(LegacyOrder.EDIT_WIEDERVORLAGE).build();

			var translatedCommand = processor.translateUpdateAttachedItem(editWiedervorlageCommand);

			assertThat(translatedCommand).usingRecursiveComparison().isEqualTo(expectedCommand);
		}

		@Test
		void shouldNotTranslateOtherOrderCommands() {
			var command = editWiedervorlageCommand.toBuilder().order(LoremIpsum.getInstance().getWords(1)).build();

			var translatedCommand = processor.translateUpdateAttachedItem(command);

			assertThat(translatedCommand).isSameAs(command);
		}

		@Test
		void shouldNotTranslateOtherAttachedItems() {
			var otherCommandBody = Map.of(CommandBodyMapper.ITEM_NAME_PROPERTY, LoremIpsum.getInstance().getWords(1));
			var command = editWiedervorlageCommand.toBuilder().body(otherCommandBody).build();

			var translatedCommand = processor.translateUpdateAttachedItem(command);

			assertThat(translatedCommand).isSameAs(command);
		}
	}

	@Nested
	class TestTranslatePatchAttachedItem {

		private final Map<String, ?> commandBody = Map.of(CommandBodyMapper.ITEM_NAME_PROPERTY, Wiedervorlage.class.getSimpleName());
		private final Command patchWiedervorlageCommand = CommandTestFactory.createBuilder()
				.order(CommandOrder.PATCH_ATTACHED_ITEM.name())
				.body(commandBody)
				.build();

		@Test
		void shouldCallTranslatePatchWiedervorlage() {
			processor.translatePatchAttachedItem(patchWiedervorlageCommand);

			verify(processor).translatePatchWiedervorlage(patchWiedervorlageCommand);
		}

		@Test
		void shouldReturnTranslatedCommand() {
			var translatedCommand = CommandTestFactory.create();
			doReturn(translatedCommand).when(processor).translatePatchWiedervorlage(any());

			var resultCommand = processor.translatePatchAttachedItem(patchWiedervorlageCommand);

			assertThat(resultCommand).isSameAs(translatedCommand);
		}

		@Test
		void shouldNotTranslateOtherOrderCommands() {
			var command = patchWiedervorlageCommand.toBuilder().order(LoremIpsum.getInstance().getWords(1)).build();

			var translatedCommand = processor.translatePatchAttachedItem(command);

			assertThat(translatedCommand).isSameAs(command);
		}

		@Test
		void shouldNotTranslateOtherAttachedItems() {
			var otherCommandBody = Map.of(CommandBodyMapper.ITEM_NAME_PROPERTY, LoremIpsum.getInstance().getWords(1));
			var command = patchWiedervorlageCommand.toBuilder().body(otherCommandBody).build();

			var translatedCommand = processor.translatePatchAttachedItem(command);

			assertThat(translatedCommand).isSameAs(command);
		}
	}

	@Nested
	class TestTranslatePatchWiedervorlage {

		@Test
		void shouldReturnWiedervorlageErledigenCommand() {
			var commandBody = Map.of(CommandBodyMapper.ITEM_PROPERTY, Map.of(Wiedervorlage.DONE_FIELD, true));
			var command = CommandTestFactory.createBuilder().body(commandBody).build();
			var expectedCommand = command.toBuilder().order(LegacyOrder.WIEDERVORLAGE_ERLEDIGEN);

			var translatedCommand = processor.translatePatchWiedervorlage(command);

			assertThat(translatedCommand).usingRecursiveComparison().isEqualTo(expectedCommand);
		}

		@Test
		void shouldReturnWiedervorlageWiedereroeffnenCommand() {
			var commandBody = Map.of(CommandBodyMapper.ITEM_PROPERTY, Map.of(Wiedervorlage.DONE_FIELD, false));
			var command = CommandTestFactory.createBuilder().body(commandBody).build();
			var expectedCommand = command.toBuilder().order(LegacyOrder.WIEDERVORLAGE_WIEDEREROEFFNEN);

			var translatedCommand = processor.translatePatchWiedervorlage(command);

			assertThat(translatedCommand).usingRecursiveComparison().isEqualTo(expectedCommand);
		}

		@Test
		void shouldNotTranslateOnMissingBoolean() {
			var commandBody = Map.of(CommandBodyMapper.ITEM_PROPERTY, Collections.emptyMap());
			var command = CommandTestFactory.createBuilder().body(commandBody).build();

			var translatedCommand = processor.translatePatchWiedervorlage(command);

			assertThat(translatedCommand).isSameAs(command);
		}
	}
}
